using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Reflection;

namespace InfiniTec.DirectoryServices
{
    /// <summary>
    /// A base class containing convinience methods for accessing ActiveDirectory
    /// entries.
    /// </summary>
    public abstract class ActiveDirectoryEntry
    {
        /// <summary>
        /// This event is triggered, when a refresh operation completes.
        /// </summary>
        public event AsyncCompletedEventHandler RefreshCompleted
        {
            add { BaseItem.RefreshCompleted += value; }
            remove { BaseItem.RefreshCompleted -= value; }
        }

        private Item _BaseItem;
        private List<string> _RequiredFields;

        private ActiveDirectoryEntry()
        {
            _RequiredFields = new List<string>();
            AddRequiredFields(SchemaMapping.CommonSchema);
        }
        /// <summary>
        /// Initializes a new instance of the <see cref="ActiveDirectoryPrincipal"/> class.
        /// </summary>
        /// <param name="distinguishedName">The distinguished Name.</param>
        /// <param name="connection">The connection.</param>
        public ActiveDirectoryEntry(string distinguishedName, Connection connection): this()
        {
            _BaseItem = new Item(distinguishedName, connection);
        }

        /// <summary>
        /// Adds the required fields which are needed to access the various properties of the
        /// object.
        /// </summary>
        /// <param name="fields">The required fields.</param>
        protected void AddRequiredFields(string[] fields)
        {
            _RequiredFields.AddRange(fields);
        }
       

        /// <summary>
        /// Initializes a new instance of the <see cref="ActiveDirectoryEntry"/> class.
        /// </summary>
        /// <param name="objectId">The object id.</param>
        /// <param name="connection">The connection.</param>
        public ActiveDirectoryEntry(Guid objectId, Connection connection): this()
        {
            _BaseItem = new Item(string.Format("<Guid={0}>", objectId), connection);
        }

        /// <summary>
        /// Provides access to the properties of the base item.
        /// </summary>
        /// <value>The properties.</value>
        public PropertyCollection Properties
        {
            [DebuggerStepThrough]
            get { return BaseItem.Properties; }
        }
	
        /// <summary>
        /// Initializes a new instance of the <see cref="ActiveDirectoryEntry"/> class.
        /// </summary>
        /// <param name="baseItem">The base item.</param>
        public ActiveDirectoryEntry(Item baseItem)
        {
            _BaseItem = baseItem;
        }

        /// <summary>
        /// Gets the base item.
        /// </summary>
        /// <value>The base item.</value>
        protected Item BaseItem
        {
            [DebuggerStepThrough]
            get { return _BaseItem; }
        }
        /// <summary>
        /// Gets the distinguished name.
        /// </summary>
        /// <value>The distinguished name.</value>
        public string DistinguishedName
        {
            get { return GetPropertyValue<string>(Constants.Common.DistinguishedName); }
        }

        /// <summary>
        /// Gets the canonicalname.
        /// </summary>
        /// <value>The canonical name.</value>
        public string CanonicalName
        {
            get
            {
                return GetPropertyValue<string>(Constants.Common.CanonicalName);
            }
        }


        /// <summary>
        /// Gets or sets the relative name of the item.
        /// </summary>
        /// <value>The relative name of the item.</value>
        public string RelativeName
        {
            get { return GetPropertyValue<string>(Constants.Common.RelativeName); }
            set { SetPropertyValue<string>(Constants.Common.RelativeName, value); }
        }

        /// <summary>
        /// Gets or sets the displayname of the item.
        /// </summary>
        /// <value>The displayname of the item.</value>
        public string DisplayName
        {
            get
            {
                return GetPropertyValue<string>(Constants.Common.DisplayName);
            }
            set
            {
                SetPropertyValue<string>(Constants.Common.DisplayName, value);
            }
        }

        /// <summary>
        /// Gets the object category.
        /// </summary>
        /// <value>The object category.</value>
        public string ObjectCategory
        {
            get
            {
                return GetPropertyValue<string>(Constants.Common.ObjectCategory);
            }
        }
        /// <summary>
        /// Gets the object classes.
        /// </summary>
        /// <value>The object class.</value>
        public IList<string> ObjectClass
        {
            get
            {
                return GetMultivaluedPropertyValue<string>(Constants.Common.ObjectClass);
            }
        }

        /// <summary>
        /// Gets the object GUID.
        /// </summary>
        /// <value>The object GUID.</value>
        public Guid ObjectGuid
        {
            get
            {
                return GetPropertyValue<Guid>(Constants.Common.ObjectGuid);
            }
        }


        /// <summary>
        /// This method is called before the <see cref="Refresh"/> or <see cref="RefreshAsync"/>
        /// operation is called.
        /// </summary>
        protected virtual void PreRefresh()
        {
            foreach (string property in _RequiredFields)
            {
                if (!BaseItem.PropertiesToLoad.Contains(property))
                    BaseItem.PropertiesToLoad.Add(property);
            }
        }
        /// <summary>
        /// This method is called before the <see cref="Save"/> or <see cref="SaveAsync"/>
        /// operation is called.
        /// </summary>
        protected virtual void PreSave()
        {
        }

        /// <summary>
        /// Synchronously refreshes the properties of the curren item.
        /// </summary>
        public void Refresh()
        {
            PreRefresh();
            BaseItem.Refresh();
        }
        /// <summary>
        /// Asynchronously refreshes the properties of the curren item.
        /// </summary>
        public void RefreshAsync()
        {
            PreRefresh();
            BaseItem.RefreshAsync();
        }
        /// <summary>
        /// Synchronously saves the properties of the curren item.
        /// </summary>
        public void Save()
        {
            PreSave();

            throw new NotImplementedException();
//            BaseItem.Save();
        }
        /// <summary>
        /// Asynchronously saves the properties of the curren item.
        /// </summary>
        public void SaveAsync()
        {
            PreSave();
            throw new NotImplementedException();
            //            BaseItem.SaveAsync();
        }
        /// <summary>
        /// Synchronously deletes the properties of the curren item.
        /// </summary>
        public void Delete()
        {
            throw new NotImplementedException();
//            BaseItem.Delete();
        }
        /// <summary>
        /// Asynchronously deletes the properties of the curren item.
        /// </summary>
        public void DeleteAsync()
        {
            throw new NotImplementedException();
//            BaseItem.DeleteAsync();
        }

        /// <summary>
        /// Gets the property value.
        /// </summary>
        /// <param name="name">The name of the property.</param>
        /// <typeparam name="T">The type of the requested property.</typeparam>
        /// <returns>The value of the requested property. If not found, the default
        /// value for the specified type is returned.</returns>
        protected T GetPropertyValue<T>(string name)
        {
            return GetPropertyValue<T>(name, default(T));
        }
        /// <summary>
        /// Gets the property value.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <typeparam name="T">The type of the requested property.</typeparam>
        /// <returns>The value of the requested property. If not found, the default
        /// value for the specified type is returned.</returns>
        protected T? GetNullablePropertyValue<T>(string name)
            where T : struct
        {
            Property<T> prop;

            if (!BaseItem.Properties.TryGetProperty<T>(name, out prop)) return null;
            else return prop.Value;
        }
        /// <summary>
        /// Gets the multi-valued property value.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <typeparam name="T">The type of the requested property.</typeparam>
        /// <returns>The value of the requested property. If not found, the default
        /// value for the specified type is returned.</returns>
        protected IList<T> GetMultivaluedPropertyValue<T>(string name)
        {
            return GetMultivaluedPropertyValue<T>(name, null);
        }
        /// <summary>
        /// Gets the value of the property.
        /// </summary>
        /// <param name="name">The name of the property.</param>
        /// <param name="defaultValue">The default value for the property</param>
        /// <typeparam name="T">The type of the requested property.</typeparam>
        /// <returns>The value of the requested property. If not found, specified
        /// default value is returned.</returns>
        protected T GetPropertyValue<T>(string name, T defaultValue)
        {
            Property<T> prop;

            if (!BaseItem.Properties.TryGetProperty<T>(name, out prop)) return defaultValue;
            else return prop.Value;
        }
        /// <summary>
        /// Gets the property value.
        /// </summary>
        /// <param name="name">The name of the property.</param>
        /// <param name="defaultValue">The default value for the property</param>
        /// <typeparam name="T">The type of the requested property.</typeparam>
        /// <returns>The value of the requested property. If not found, the specified default
        /// value is returned.</returns>
        protected IList<T> GetMultivaluedPropertyValue<T>(string name, T[] defaultValue)
        {
            MultiValuedProperty<T> prop;

            if (!BaseItem.Properties.TryGetMultiValuedProperty<T>(name, out prop)) return defaultValue;
            else return prop.Values;
        }
        /// <summary>
        /// Gets the property value.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="value">The vlaue of the value.</param>
        /// <typeparam name="T">The type of the requested property.</typeparam>
        /// <returns>The value of the requested property. If not found, the default
        /// value for the specified type is returned.</returns>
        protected void SetPropertyValue<T>(string name, T value)
        {
            Property<T> prop;

            if (value == null) BaseItem.Properties.Remove(name);

            if (!BaseItem.Properties.TryGetProperty<T>(name, out prop))
            {
                prop = new Property<T>(name, value);
                BaseItem.Properties.Add(prop);
            }
            else
            {
                prop.Value = value;
            }
        }

        /// <summary>
        /// Sets the nullable property value.
        /// </summary>
        /// <typeparam name="T">The type of the property</typeparam>
        /// <param name="name">The name of the property.</param>
        /// <param name="value">The value of the property.</param>
        protected void SetNullablePropertyValue<T>(string name, T? value)
            where T : struct
        {
            Property<T> prop;

            if (value == null) BaseItem.Properties.Remove(name);

            if (!BaseItem.Properties.TryGetProperty<T>(name, out prop))
            {
                prop = new Property<T>(name, value.Value);
                BaseItem.Properties.Add(prop);
            }
            else
            {
                prop.Value = value.Value;
            }
        }


        /// <summary>
        /// Ensures that the property with the specified name and type is present in the 
        /// and can be accessed.
        /// </summary>
        /// <param name="name">The name of the property.</param>
        protected void EnsureMultiValuedProperty<T>(string name)
        {
            if (Properties.Contains(name)) return;

            Properties.AddMultiValued<T>(name, (T[])null);
        }

        /// <summary>
        /// Returns wheter the specified flags are set in the given value.
        /// </summary>
        /// <param name="value">The value to check.</param>
        /// <param name="flags">The flags to check</param>
        /// <returns>Returns true, if the flags are set in the specified value. False otherwise. If 
        /// the specified value is null, null is returned.</returns>
        protected bool? CheckFlagValue(int? value, int flags)
        {
            if (value == null) return null;

            return ((value.Value & flags) == flags);
        }

        [MethodImpl(MethodImplOptions.Synchronized)]
        private void EnsureMultivaluedPropertyExists<T>(string name)
        {
            if (BaseItem.Properties.Contains(name)) return;

            BaseItem.Properties.Add(new MultiValuedProperty<T>(name));
        }

        /// <summary>
        /// Gets the schema entry for the current item.
        /// </summary>
        /// <returns></returns>
        public Item GetSchemaEntry()
        {
            return new Item(ObjectCategory, BaseItem.Connection);
        }

    }
}
