using System;
using System.Collections.Generic;
using System.Text;
using System.Security.Principal;

namespace InfiniTec.DirectoryServices
{
    /// <summary>
    /// Extends the <see cref="ActiveDirectoryPrincipal"/> class with properties
    /// unique to ActiveDirectory users.
    /// </summary>
    public class ActiveDirectoryUser: ActiveDirectoryPrincipal
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="ActiveDirectoryUser"/> class.
        /// </summary>
        /// <param name="distinguishedName">The distinguished Name.</param>
        /// <param name="connection">The connection.</param>
        public ActiveDirectoryUser(string distinguishedName, Connection connection):
            base(distinguishedName, connection)
        {
            AddRequiredFields(SchemaMapping.UserSchema);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ActiveDirectoryUser"/> class.
        /// </summary>
        /// <param name="objectId">The object id.</param>
        /// <param name="connection">The connection.</param>
        public ActiveDirectoryUser(Guid objectId, Connection connection):
            base(objectId, connection)
        {
            AddRequiredFields(SchemaMapping.UserSchema);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ActiveDirectoryUser"/> class.
        /// </summary>
        /// <param name="identifier">The identifier.</param>
        /// <param name="connection">The connection.</param>
        public ActiveDirectoryUser(SecurityIdentifier identifier, Connection connection):
            base(identifier, connection) {}

        /// <summary>
        /// Initializes a new instance of the <see cref="ActiveDirectoryUser"/> class.
        /// </summary>
        /// <param name="baseItem">The base item.</param>
        public ActiveDirectoryUser(Item baseItem): base(baseItem)
        {
        }

        /// <summary>
        /// Gets or sets the country.
        /// </summary>
        /// <value>The country.</value>
        public string Country
        {
            get { return GetPropertyValue<string>(Constants.User.Country); }
            set { SetPropertyValue<string>(Constants.User.Country, value); }
        }
        /// <summary>
        /// Gets or sets the company.
        /// </summary>
        /// <value>The company.</value>
        public string Company
        {
            get { return GetPropertyValue<string>(Constants.User.Company); }
            set { SetPropertyValue<string>(Constants.User.Company, value); }
        }
        /// <summary>
        /// Gets or sets the department.
        /// </summary>
        /// <value>The department.</value>
        public string Department
        {
            get { return GetPropertyValue<string>(Constants.User.Department); }
            set { SetPropertyValue<string>(Constants.User.Department, value); }
        }
        /// <summary>
        /// Gets or sets the description.
        /// </summary>
        /// <value>The description.</value>
        public string Description
        {
            get { return GetPropertyValue<string>(Constants.User.Description); }
            set { SetPropertyValue<string>(Constants.User.Description, value); }
        }

        /// <summary>
        /// Gets or sets the fax number.
        /// </summary>
        /// <value>The fax number.</value>
        public string FaxNumber
        {
            get { return GetPropertyValue<string>(Constants.User.FaxNumber); }
            set { SetPropertyValue<string>(Constants.User.FaxNumber, value); }
        }

        /// <summary>
        /// Gets or sets the first name.
        /// </summary>
        /// <value>The name of the first.</value>
        public string FirstName
        {
            get { return GetPropertyValue<string>(Constants.User.FirstName); }
            set { SetPropertyValue<string>(Constants.User.FirstName, value); }
        }

        /// <summary>
        /// Gets or sets the home phone.
        /// </summary>
        /// <value>The home phone.</value>
        public string HomePhone
        {
            get { return GetPropertyValue<string>(Constants.User.HomePhone); }
            set { SetPropertyValue<string>(Constants.User.HomePhone, value); }
        }

        /// <summary>
        /// Gets or sets the initials.
        /// </summary>
        /// <value>The initials.</value>
        public string Initials
        {
            get { return GetPropertyValue<string>(Constants.User.Initials); }
            set { SetPropertyValue<string>(Constants.User.Initials, value); }
        }

        /// <summary>
        /// Gets or sets the IP phone.
        /// </summary>
        /// <value>The IP phone.</value>
        public string IPPhone
        {
            get { return GetPropertyValue<string>(Constants.User.IPPhone); }
            set { SetPropertyValue<string>(Constants.User.IPPhone, value); }
        }

        /// <summary>
        /// Gets or sets the city.
        /// </summary>
        /// <value>The city.</value>
        public string City
        {
            get { return GetPropertyValue<string>(Constants.User.City); }
            set { SetPropertyValue<string>(Constants.User.City, value); }
        }

        /// <summary>
        /// Gets or sets the mail address.
        /// </summary>
        /// <value>The mail address.</value>
        public string MailAddress
        {
            get { return GetPropertyValue<string>(Constants.User.MailAddress); }
            set { SetPropertyValue<string>(Constants.User.MailAddress, value); }
        }

        /// <summary>
        /// Gets or sets the manager.
        /// </summary>
        /// <value>The manager.</value>
        public string Manager
        {
            get { return GetPropertyValue<string>(Constants.User.Manager); }
            set { SetPropertyValue<string>(Constants.User.Manager, value); }
        }

        /// <summary>
        /// Gets or sets the mobile phone.
        /// </summary>
        /// <value>The mobile phone.</value>
        public string MobilePhone
        {
            get { return GetPropertyValue<string>(Constants.User.Mobile); }
            set { SetPropertyValue<string>(Constants.User.Mobile, value); }
        }

        /// <summary>
        /// Gets the other fax numbers.
        /// </summary>
        /// <value>The other fax numbers.</value>
        public IList<string> OtherFaxNumbers
        {
            get
            {
                EnsureMultiValuedProperty<string>(Constants.User.OtherFaxNumbers);
                return GetMultivaluedPropertyValue<string>(Constants.User.OtherFaxNumbers);
            }
        }

        /// <summary>
        /// Gets the other home numbers.
        /// </summary>
        /// <value>The other home numbers.</value>
        public IList<string> OtherHomeNumbers
        {
            get
            {
                EnsureMultiValuedProperty<string>(Constants.User.OtherHomePhones);
                return GetMultivaluedPropertyValue<string>(Constants.User.OtherHomePhones);
            }
        }

        /// <summary>
        /// Gets the other IP phone numbers.
        /// </summary>
        /// <value>The other IP phone numbers.</value>
        public IList<string> OtherIPPhoneNumbers
        {
            get
            {
                EnsureMultiValuedProperty<string>(Constants.User.OtherIPPhones);
                return GetMultivaluedPropertyValue<string>(Constants.User.OtherIPPhones);
            }
        }

        /// <summary>
        /// Gets the other mobile phone numbers.
        /// </summary>
        /// <value>The other mobile phone numbers.</value>
        public IList<string> OtherMobilePhoneNumbers
        {
            get
            {
                EnsureMultiValuedProperty<string>(Constants.User.OtherMobile);
                return GetMultivaluedPropertyValue<string>(Constants.User.OtherMobile);
            }
        }
        /// <summary>
        /// Gets the other pager numbers.
        /// </summary>
        /// <value>The other pager numbers.</value>
        public IList<string> OtherPagerNumbers
        {
            get
            {
                EnsureMultiValuedProperty<string>(Constants.User.OtherPager);
                return GetMultivaluedPropertyValue<string>(Constants.User.OtherPager);
            }
        }

        /// <summary>
        /// Gets the other telephone numbers.
        /// </summary>
        /// <value>The other telephone numbers.</value>
        public IList<string> OtherTelephoneNumbers
        {
            get
            {
                EnsureMultiValuedProperty<string>(Constants.User.OtherTelephoneNumbers);
                return GetMultivaluedPropertyValue<string>(Constants.User.OtherTelephoneNumbers);
            }
        }

        /// <summary>
        /// Gets or sets the pager number.
        /// </summary>
        /// <value>The pager number.</value>
        public string PagerNumber
        {
            get { return GetPropertyValue<string>(Constants.User.PagerNumber); }
            set { SetPropertyValue<string>(Constants.User.PagerNumber, value); }
        }

        /// <summary>
        /// Gets or sets the office.
        /// </summary>
        /// <value>The office.</value>
        public string Office
        {
            get { return GetPropertyValue<string>(Constants.User.Office); }
            set { SetPropertyValue<string>(Constants.User.Office, value); }
        }

        /// <summary>
        /// Gets or sets the postal code.
        /// </summary>
        /// <value>The postal code.</value>
        public string PostalCode
        {
            get { return GetPropertyValue<string>(Constants.User.PostalCode); }
            set { SetPropertyValue<string>(Constants.User.PostalCode, value); }
        }

        /// <summary>
        /// Gets or sets the last name
        /// </summary>
        /// <value>The name of the last.</value>
        public string LastName
        {
            get { return GetPropertyValue<string>(Constants.User.LastName); }
            set { SetPropertyValue<string>(Constants.User.LastName, value); }
        }

        /// <summary>
        /// Gets or sets the post office box.
        /// </summary>
        /// <value>The post office box.</value>
        public string PostOfficeBox
        {
            get { return GetPropertyValue<string>(Constants.User.PostOfficeBox); }
            set { SetPropertyValue<string>(Constants.User.PostOfficeBox, value); }
        }

        /// <summary>
        /// Gets or sets the state.
        /// </summary>
        /// <value>The state.</value>
        public string State
        {
            get { return GetPropertyValue<string>(Constants.User.State); }
            set { SetPropertyValue<string>(Constants.User.State, value); }
        }

        /// <summary>
        /// Gets or sets the street address.
        /// </summary>
        /// <value>The street address.</value>
        public string StreetAddress
        {
            get { return GetPropertyValue<string>(Constants.User.StreetAddress); }
            set { SetPropertyValue<string>(Constants.User.StreetAddress, value); }
        }

        /// <summary>
        /// Gets or sets the telephone number.
        /// </summary>
        /// <value>The telephone number.</value>
        public string TelephoneNumber
        {
            get { return GetPropertyValue<string>(Constants.User.TelephoneNumber); }
            set { SetPropertyValue<string>(Constants.User.TelephoneNumber, value); }
        }
        /// <summary>
        /// Gets or sets the title.
        /// </summary>
        /// <value>The title.</value>
        public string Title
        {
            get { return GetPropertyValue<string>(Constants.User.Title); }
            set { SetPropertyValue<string>(Constants.User.Title, value); }
        }
        /// <summary>
        /// Gets or sets the other web page.
        /// </summary>
        /// <value>The other web page.</value>
        public string OtherWebPage
        {
            get { return GetPropertyValue<string>(Constants.User.OtherWebPage); }
            set { SetPropertyValue<string>(Constants.User.OtherWebPage, value); }
        }

        /// <summary>
        /// Gets or sets the user principal name.
        /// </summary>
        /// <value>The name of the user principal.</value>
        public string UserPrincipalName
        {
            get { return GetPropertyValue<string>(Constants.User.UserPrincipalName); }
            set { SetPropertyValue<string>(Constants.User.UserPrincipalName, value); }
        }

        /// <summary>
        /// Gets or sets the web page.
        /// </summary>
        /// <value>The web page.</value>
        public string WebPage
        {
            get { return GetPropertyValue<string>(Constants.User.WebPage); }
            set { SetPropertyValue<string>(Constants.User.WebPage, value); }
        }

        /// <summary>
        /// Gets the item of the manager of the current user.
        /// </summary>
        /// <returns></returns>
        public ActiveDirectoryUser GetManagerItem()
        {
            if (string.IsNullOrEmpty(Manager)) throw new ManagerNotSpecifiedException("No manager has been specified on item " + DistinguishedName);

            return new ActiveDirectoryUser(Manager, BaseItem.Connection);
        }

        

      
    }
}
