using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.Net;
using System.DirectoryServices.Protocols;
using InfiniTec.Threading;
using System.Threading;
using System.Runtime.CompilerServices;

namespace InfiniTec.DirectoryServices
{
    /// <summary>
    /// The Connection class provides access to an ActiveDirectory instance by using
    /// an <see cref="LdapConnection"/> instance as well as a native connection.
    /// </summary>
    public class Connection : IDisposable
    {
        private object _NativeConnectionSyncRoot = new object();
        private string _DnsDomainName;
        private NetworkCredential _Credential;
        private AuthType _AuthenticationType;
        private LdapConnection _Connection;
        private int _Initializing;
        private bool _Initialized;
        private ManualResetEvent _InitializationWaitHandle;
        private PartialResultProcessing _PartialResultProcessing;
        private DirectoryServiceConnectionSafeHandle _DirectoryConnection;
        private string _DomainController;

        /// <summary>
        /// Gets or sets the domain controller.
        /// </summary>
        /// <value>The domain controller.</value>
        public string DomainController
        {
            [DebuggerStepThrough]
            get { return _DomainController; }
            [DebuggerStepThrough]
            set { _DomainController = value; }
        }

        internal object NativeConnectionSyncRoot
        {
            get { return _NativeConnectionSyncRoot; }
        }

        internal DirectoryServiceConnectionSafeHandle DirectoryConnection
        {
            [DebuggerStepThrough]
            get
            {
                EnsureNativeDirectoryConnection();
                return _DirectoryConnection;
            }
        }

        private void EnsureNativeDirectoryConnection()
        {
            lock (_NativeConnectionSyncRoot)
            {
                if (_DirectoryConnection != null) return;

                //if (!_Initialized) Initialize();

                _DirectoryConnection = DirectoryServiceConnectionSafeHandle.Create(
                    DomainController, DnsDomainName, Credential);
            }
        }

        /// <summary>
        /// Gets the session options.
        /// </summary>
        /// <value>The session options.</value>
        public LdapSessionOptions SessionOptions
        {
            [DebuggerStepThrough]
            get { return _Connection.SessionOptions; }
        }

        /// <summary>
        /// Specifies, whether partial results are returned during a query.
        /// </summary>
        public PartialResultProcessing PartialResultProcessing
        {
            [DebuggerStepThrough]
            get { return _PartialResultProcessing; }
            [DebuggerStepThrough]
            set { _PartialResultProcessing = value; }
        }


        /// <summary>
        /// Gets or sets the <see cref="AuthType"/> which is used to connect to the directory server.
        /// </summary>
        public AuthType AuthenticationType
        {
            [DebuggerStepThrough]
            get { return _AuthenticationType; }
            [DebuggerStepThrough]
            set { _AuthenticationType = value; }
        }

        /// <summary>
        /// Gets or sets the credentials which are used to connect to the directory server.
        /// If set to null, integrated authentication is used.
        /// </summary>
        public NetworkCredential Credential
        {
            [DebuggerStepThrough]
            get { return _Credential; }
            [DebuggerStepThrough]
            set { _Credential = value; }
        }

        /// <summary>
        /// Specifies the host to bind to when performing Active Directory requests.
        /// This can bei either a computername, IP address or full qualified domain name.
        /// If set to null, a serverless binding is performed.
        /// </summary>
        public string DnsDomainName
        {
            [DebuggerStepThrough]
            get { return _DnsDomainName; }
            [DebuggerStepThrough]
            set { _DnsDomainName = value; }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Connection"/> class.
        /// </summary>
        /// <param name="domainBinding">The domain binding.</param>
        /// <param name="type">The type of the specified domain binding.</param>
        /// <param name="connectionLess">if set to <c>true</c> [connection less].</param>
        public Connection(string domainBinding, DirectoryIdentifierType type, bool connectionLess)
        {
            _InitializationWaitHandle = new ManualResetEvent(false);

            if (type == DirectoryIdentifierType.DnsDomain)
            {
                DnsDomainName = domainBinding;
            }
            else if (type == DirectoryIdentifierType.Server)
            {
                DomainController = domainBinding;
            }
            else
            {
                throw new ArgumentException("type", "The specified directory identifier type is invalid.");
            }
            _Connection = new LdapConnection(new LdapDirectoryIdentifier(domainBinding, (type == DirectoryIdentifierType.DnsDomain) ? true : false, connectionLess));
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Connection"/> class.
        /// </summary>
        public Connection() : this(null, DirectoryIdentifierType.DnsDomain, false) { }

        /// <summary>
        /// Initializes a new instance of the <see cref="Connection"/> class.
        /// </summary>
        /// <param name="domainBinding">The domain binding to use.</param>
        /// <param name="type">The type of the domain binding (either server or DNS domain name).</param>
        /// <param name="connectionLess">if set to <c>true</c>, the connection is connection-less; otherwise, the connection is kept alive.</param>
        /// <param name="credential">The credential to use.</param>
        /// <param name="authType">The type of the authentication to use.</param>
        public Connection(string domainBinding, DirectoryIdentifierType type, bool connectionLess, NetworkCredential credential, AuthType authType)
            : this(domainBinding, type, connectionLess)
        {
            Credential = credential;
            AuthenticationType = authType;
        }

        /// <summary>
        /// Gets an asynchronous operation which can be used to perform a request.
        /// </summary>
        /// <param name="request">The request to perform.</param>
        /// <returns>An async operation.</returns>
        public RequestOperation<object, TResponse> GetSendRequestOperation<TRequest, TResponse>(TRequest request)
            where TRequest : DirectoryRequest
            where TResponse : DirectoryResponse
        {
            return new RequestOperation<object, TResponse>(SendRequestInternal<TRequest, TResponse, object>(request, false, null));
        }

        /// <summary>
        /// Gets an asynchronous operation which can be used to perform a request.
        /// </summary>
        /// <param name="request">The request to perform.</param>
        /// <param name="partialResultConverter">A callback which is called to transform partial results before passing them to the progress changed event handler.</param>
        /// <returns>An async operation.</returns>
        public RequestOperation<IList<TPartialResult>, TResponse> GetSendRequestOperation<TRequest, TResponse, TPartialResult>(TRequest request, Converter<object, TPartialResult> partialResultConverter)
            where TRequest : DirectoryRequest
            where TResponse : DirectoryResponse
            where TPartialResult : class
        {
            return new RequestOperation<IList<TPartialResult>, TResponse>(SendRequestInternal<TRequest, TResponse, TPartialResult>(request, true, partialResultConverter));
        }

        private IEnumerable<OperationAction> SendRequestInternal<TRequest, TResponse, TPartialResult>(TRequest request, bool enablePartialResults, Converter<object, TPartialResult> partialResultConverter)
            where TRequest : DirectoryRequest
            where TResponse : DirectoryResponse
            where TPartialResult : class
        {
            IAsyncResult ar = null;
            TResponse response;
            PartialResultsCollection partialResult;
            LdapConnection connection = null;
            IList<TPartialResult> partialResultItems;

            try
            {
                if (!_Initialized)
                {
                    if (Interlocked.Exchange(ref _Initializing, 1) == 1)
                    {
                        _InitializationWaitHandle.WaitOne();
                        if (!_Initialized) throw new InitializationFailedException("The initialization failed during another call");
                    }
                    else
                    {
                        Initialize();
                    }
                }
                connection = _Connection;
                ar = connection.BeginSendRequest(request, enablePartialResults ? PartialResultProcessing.ReturnPartialResults : PartialResultProcessing.NoPartialResultSupport, null, null);

                yield return new ProgressUpdate<RequestProgress>(RequestProgress.PerformingRequest, ar.AsyncWaitHandle);

                if (enablePartialResults)
                {

                    while (!ar.IsCompleted)
                    {
                        partialResult = _Connection.GetPartialResults(ar);

                        partialResultItems = ConvertItems<TPartialResult>(partialResult, partialResultConverter);

                        yield return new ProgressUpdate<IList<TPartialResult>>(partialResultItems, ar.AsyncWaitHandle);
                    }
                }

                response = (TResponse)connection.EndSendRequest(ar);

                ar = null;

                yield return new OperationResult<TResponse>(response);
            }
            finally
            {
                if (ar != null && connection != null)
                {
                    connection.Abort(ar);
                }
            }
        }

        private static IList<TPartialResult> ConvertItems<TPartialResult>(PartialResultsCollection result, Converter<object, TPartialResult> partialResultConverter)
            where TPartialResult : class
        {
            List<TPartialResult> partialResultItems;
            TPartialResult resultItem;

            partialResultItems = new List<TPartialResult>();

            foreach (object entry in result)
            {
                if (partialResultConverter == null)
                {
                    resultItem = entry as TPartialResult;
                    if (resultItem == null) throw new InvalidOperationException("The partial result item cannot be cast to type " + typeof(TPartialResult).Name + " and no converter was provided");

                    partialResultItems.Add(resultItem);
                }
                else
                {
                    resultItem = partialResultConverter(entry);
                    if (resultItem == null) continue;

                    partialResultItems.Add(resultItem);
                }
            }
            return partialResultItems.AsReadOnly();
        }

        private void Initialize()
        {
            _Connection.Bind(Credential);
            _Initialized = true;
            _Initializing = 0;

            _InitializationWaitHandle.Set();
        }

        #region IDisposable Members

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        /// <param name="disposing">if set to <c>true</c>, the method is called from the <see cref="Dispose()"/> method. Otherwise, it is called
        /// from a destructor.</param>
        protected virtual void Dispose(bool disposing)
        {
            LdapConnection connection;
            DirectoryServiceConnectionSafeHandle internalConnection;

            if (!disposing) return;

            connection = Interlocked.Exchange<LdapConnection>(ref _Connection, null);

            lock (_NativeConnectionSyncRoot)
            {
                internalConnection = Interlocked.Exchange<DirectoryServiceConnectionSafeHandle>(ref _DirectoryConnection, null);
            }

            if (internalConnection != null) internalConnection.Dispose();
            if (connection != null) connection.Dispose();
        }

        #endregion
    }
}

