using System;
using System.Collections.Generic;
using System.Text;
using System.DirectoryServices.Protocols;
using System.Diagnostics;
using System.ComponentModel;
using Exp = InfiniTec.Expressions.Expression;
using InfiniTec.Expressions;
using InfiniTec.Threading;

namespace InfiniTec.DirectoryServices
{
    /// <summary>
    /// This class is a decorator for the <see cref="Searcher"/> class which can be used
    /// to search for ActiveDirectory principals: Users, groups and contacts.
    /// </summary>
    public class PrincipalResolver : AsyncOperationObject
    {
        /// <summary>
        /// This event is invoked when the the search operation has completed.
        /// </summary>
        public event EventHandler<AsyncCompletedEventArgs> FindCompleted;

        private int _SizeLimit;
        private List<string> _PropertiesToLoad;
        private Connection _Connection;
        private string _SearchRoot;
        private int _PageSize;
        private List<SortKey> _SortKeys;
        private ResolveScope _ResolveScope;
        private SearchResultCollection _SearchResult;

        /// <summary>
        /// Gets or sets the resolve scope.
        /// </summary>
        /// <value>The resolve scope.</value>
        public ResolveScope ResolveScope
        {
            [DebuggerStepThrough]
            get { return _ResolveScope; }
            [DebuggerStepThrough]
            set { _ResolveScope = value; }
        }

        /// <summary>
        /// Gets the sort keys.
        /// </summary>
        /// <value>The sort keys.</value>
        public IList<SortKey> SortKeys
        {
            [DebuggerStepThrough]
            get { return _SortKeys; }
        }

        /// <summary>
        /// Gets or sets the size of the page.
        /// </summary>
        /// <value>The size of the page.</value>
        public int PageSize
        {
            [DebuggerStepThrough]
            get { return _PageSize; }
            [DebuggerStepThrough]
            set { _PageSize = value; }
        }

        /// <summary>
        /// Gets or sets the search root.
        /// </summary>
        /// <value>The search root.</value>
        public string SearchRoot
        {
            [DebuggerStepThrough]
            get { return _SearchRoot; }
            [DebuggerStepThrough]
            set { _SearchRoot = value; }
        }

        /// <summary>
        /// Gets the connection to use.
        /// </summary>
        /// <value>The connection to use.</value>
        public Connection Connection
        {
            [DebuggerStepThrough]
            get { return _Connection; }
        }

        /// <summary>
        /// Gets the properties to load.
        /// </summary>
        /// <value>The properties to load.</value>
        public IList<string> PropertiesToLoad
        {
            [DebuggerStepThrough]
            get { return _PropertiesToLoad; }
        }

        /// <summary>
        /// Gets or sets the size limit.
        /// </summary>
        /// <remarks>
        /// If more entries are found than specified by this value, an exception is thrown.
        /// </remarks>
        /// <value>The size limit.</value>
        public int SizeLimit
        {
            [DebuggerStepThrough]
            get { return _SizeLimit; }
            [DebuggerStepThrough]
            set { _SizeLimit = value; }
        }

        /// <summary>
        /// Gets the search result.
        /// </summary>
        /// <value>The search result.</value>
        public IList<ActiveDirectoryPrincipal> SearchResult
        {
            [DebuggerStepThrough]
            get { return _SearchResult; }
        }


        /// <summary>
        /// Initializes a new instance of the <see cref="PrincipalResolver"/> class.
        /// </summary>
        /// <param name="connection">The connection to use.</param>
        public PrincipalResolver(Connection connection)
            : this(null, connection)
        {
        }
        /// <summary>
        /// Initializes a new instance of the <see cref="PrincipalResolver"/> class.
        /// </summary>
        /// <param name="searchRoot">The search root.</param>
        /// <param name="connection">The connection to use.</param>
        public PrincipalResolver(string searchRoot, Connection connection)
        {
            ResolveScope = ResolveScope.Forest;
            _Connection = connection;

            _SearchRoot = searchRoot;
            _PropertiesToLoad = new List<string>();
            _SortKeys = new List<SortKey>();
            _SearchResult = new SearchResultCollection();
        }


        /// <summary>
        /// Calls the <see cref="M:ProgressCompleted"/> event.
        /// </summary>
        /// <param name="e"></param>
        protected virtual void OnCompleted(AsyncCompletedEventArgs e)
        {
            if (FindCompleted != null) FindCompleted(this, e);
        }

        /// <summary>
        /// Synchronously starts the search operation.
        /// </summary>
        /// <returns></returns>
        public void FindAll(string name, ResolveTypes types)
        {
            using (AsyncOperation<IList<Item>> op = GetFindOperation(name, types, SearchMode.FindAll))
            {
                op.Run();
                op.WaitForCompletion();
            }
        }
        /// <summary>
        /// Synchronously starts the search operation, finding the next page in the search
        /// </summary>
        /// <returns></returns>
        public void FindPage(string name, ResolveTypes types)
        {
            using (AsyncOperation<IList<Item>> op = GetFindOperation(name, types, SearchMode.FindPage))
            {
                op.Run();
                op.WaitForCompletion();
            }
        }

        private void FindAsync(string name, ResolveTypes types, SearchMode mode)
        {
            AsyncOperation<IList<Item>> operation;

            InitiatePendingAsyncOperation();

            operation = GetFindOperation(name, types, mode);

            RegisterPendingAsyncOperation(operation);

            operation.Completed += SearchAsyncCompleted;
            operation.Error += SearchAsyncError;

            operation.Run();
        }

        /// <summary>
        /// Asynchronously starts the search operation.
        /// </summary>
        /// <param name="name">The name to search for</param>
        /// <param name="types">The search types</param>
        public void FindAllAsync(string name, ResolveTypes types)
        {
            FindAsync(name, types, SearchMode.FindAll);
        }

        /// <summary>
        /// Asynchronously starts the search operation.
        /// </summary>
        public void FindPageAsync(string name, ResolveTypes types)
        {
            FindAsync(name, types, SearchMode.FindPage);
        }

        void SearchAsyncError(object sender, ExceptionEventArgs e)
        {
            UnregisterPendingAsyncOperation();
            OnCompleted(new AsyncCompletedEventArgs(e.Exception, false, null));
        }

        void SearchAsyncCompleted(object sender, CompletedEventArgs e)
        {
            UnregisterPendingAsyncOperation();

            if (e.Status == OperationStatus.Completed)
            {
                OnCompleted(new AsyncCompletedEventArgs(null, false, null));
            }
            else
            {
                OnCompleted(new AsyncCompletedEventArgs(null, true, null));
            }
        }

        /// <summary>
        /// Returns the search operation used to perform the search based
        /// ojn the specified settings.
        /// </summary>
        /// <returns></returns>
        public virtual AsyncOperation<IList<Item>> GetFindOperation(string name, ResolveTypes types, SearchMode mode)
        {
            return new AsyncOperation<IList<Item>>(FindOperationInternal(name, types, mode));
        }

        private IEnumerable<OperationAction> FindOperationInternal(string name, ResolveTypes types, SearchMode mode)
        {
            Searcher searcher;
            ActiveDirectoryPrincipal entry;


            if (types == ResolveTypes.None)
                throw new ArgumentException("At least one type of principals must be searched.");

            searcher = new Searcher(SearchRoot, Connection, _PropertiesToLoad.ToArray());
            searcher.Scope = SearchScope.Subtree;
            searcher.IncludeDeletedItems = false;
            searcher.PageSize = PageSize;

            foreach (string property in GetPropertiesToLoad(types))
            {
                searcher.PropertiesToLoad.Add(property);
            }

            switch (ResolveScope)
            {
                case ResolveScope.Domain: searcher.NamingContextScope = NamingContextScope.Domain; break;
                case ResolveScope.Forest: searcher.NamingContextScope = NamingContextScope.IncludeSubDomains; break;
                default: throw new ArgumentException("ResolveScope", "The specified resolve scope is invalid.");
            }

            searcher.SizeLimit = SizeLimit;
            foreach (SortKey key in SortKeys)
            {
                searcher.SortKeys.Add(key);
            }

            _SearchResult.Clear();

            searcher.Constraints = CreateLdapFilter(name, types);
            searcher.ProgressChanged += Searcher_ProgressChanged;

            using (AsyncOperation<IList<Item>> operation = searcher.GetFindOperation(mode, false))
            {
                yield return new NestedOperation(operation, false);
            }

            foreach (Item item in searcher.SearchResult)
            {
                entry = ParseSearchResultItem(item);
                if (_SearchResult.Contains(entry)) continue;
                _SearchResult.Add(entry);
            }

            searcher.SearchResult.ToString();
        }

        private IEnumerable<string> GetPropertiesToLoad(ResolveTypes scope)
        {
            foreach (string property in PropertiesToLoad)
            {
                yield return property;
            }
            foreach (string property in SchemaMapping.CommonSchema)
            {
                yield return property;
            }

            // Since tokenGroups and other computed groupmemberships cannot be programmatically
            // fetched, they are ommited here.
            yield return Constants.GroupMembership.MemberOf;

            if ((scope & ResolveTypes.User) == ResolveTypes.User)
            {
                foreach (string property in SchemaMapping.UserSchema)
                {
                    yield return property;
                }
            }
            if ((scope & ResolveTypes.Groups) == ResolveTypes.Groups)
            {
                foreach (string property in SchemaMapping.GroupSchema)
                {
                    yield return property;
                }
            }
        }

        void Searcher_ProgressChanged(object sender, SearchProgressChangedEventArgs e)
        {
            foreach (Item item in e.Items)
            {
                _SearchResult.Add(ParseSearchResultItem(item));
            }
        }

        private ActiveDirectoryPrincipal ParseSearchResultItem(Item item)
        {
            ActiveDirectoryPrincipal entry;
            entry = new ActiveDirectoryPrincipal(item);
            if (entry.ObjectClass.Contains(Constants.ObjectClasses.Group))
            {
                return new ActiveDirectoryGroup(item);
            }
            else
            {
                return new ActiveDirectoryUser(item);
            }
        }


        private string CreateLdapFilter(string name, ResolveTypes types)
        {
            Expression expression;

            if (string.IsNullOrEmpty(name)) throw new ArgumentNullException("name");


            expression =
                Exp.And(
                    CreateResolveTypesFilter(types),
                    Exp.Or(
                        Exp.Equals<string>(Constants.Special.AmbiguousNameResolution, name + "*"),
                        CreateInitialsExpression(name)));

            //expression = Exp.And(
            //    CreateResolveTypesFilter(types),
            //    Exp.Equals<string>(NamedProperties.Special.ANR, name + "*"));

            return expression.ToString(new LdapFormatter());
        }

        private Expression CreateResolveTypesFilter(ResolveTypes types)
        {
            List<Expression> result;

            result = new List<Expression>(2);
            if ((types & ResolveTypes.Groups) == ResolveTypes.Groups)
                result.Add(Exp.Equals<string>(Constants.Common.ObjectCategory, Constants.ObjectCategories.Group));
            if ((types & ResolveTypes.User) == ResolveTypes.User)
                result.Add(Exp.Equals<string>(Constants.Common.ObjectCategory, Constants.ObjectCategories.Person));

            return (result.Count > 1) ? Exp.Or(result.ToArray()) : result[0];
        }

        private Expression CreateInitialsExpression(string name)
        {
           if (name.Length == 2) return Exp.And(
                Exp.Equals<string>(Constants.User.FirstName, name[0] + "*"),
                Exp.Equals<string>(Constants.User.LastName, name[1] + "*"));
            else return new NullExpression();
        }

      
    }
}
