using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Net;

namespace InfiniTec.DirectoryServices
{
    /// <summary>
    /// This class manages the various properties of an <see cref="Item"/>
    /// </summary>
    public class PropertyCollection : ICollection<Property>
    {
        Dictionary<string, Property> _Properties;
        List<string> _RemovedProperties;

        internal IList<string> RemovedProperties
        {
            get { return _RemovedProperties; }
        }

        /// <summary>
        /// Creates a new instance of this class.
        /// </summary>
        public PropertyCollection()
        {
            _Properties = new Dictionary<string, Property>();
            _RemovedProperties = new List<string>();
        }

        /// <summary>
        /// Adds the property to the collection.
        /// </summary>
        /// <param name="property">The property to add</param>
        public void Add(Property property)
        {
            _Properties.Add(property.Name, property);
            if (_RemovedProperties.Contains(property.Name)) {
                _RemovedProperties.Remove(property.Name);
            }
        }

        /// <summary>
        /// Adds the specified properties to the collection.
        /// </summary>
        /// <param name="properties">The properties to add</param>
        public void AddRange(IEnumerable<Property> properties)
        {
            foreach (Property property in properties)
            {
                Add(property);
            }
        }

       
        /// <summary>
        /// Creates a new <see cref="Property"/> and adds it to the collection.
        /// </summary>
        /// <typeparam name="T">The type of the property</typeparam>
        /// <param name="name">The name of the property, 
        /// including the namespace and the local name.</param>
        /// <param name="value">The value of the property</param>
        public void Add<T>(string name, T value)
        {
            Add(new Property<T>(name, value));
        }

        /// <summary>
        /// Creates a new <see cref="Property"/> and adds it to the collection.
        /// </summary>
        /// <typeparam name="T">The type of the property</typeparam>
        /// <param name="name">The name of the property, 
        /// including the namespace and the local name.</param>
        /// <param name="values">The value of the property</param>
        public void AddMultiValued<T>(string name, params T[] values)
        {
            Add(new MultiValuedProperty<T>(name, values));
        }
        /// <summary>
        /// Returns true, if the collection contains a <see cref="Property"/>
        /// with the same name as the specified <see cref="Property"/>
        /// </summary>
        /// <param name="item">The <see cref="Property"/> to search for.</param>
        /// <returns>True, if the collection contains another 
        /// property with the same name.</returns>
        public bool Contains(Property item)
        {
            return Contains(item.Name);
        }

        /// <summary>
        /// Returns true, if the collection contains a <see cref="Property"/>
        /// with the specified name
        /// </summary>
        /// <param name="name">The name of the property to search for.</param>
        /// <returns>True, if the collection contains another 
        /// property with the same name.</returns>
        public bool Contains(string name)
        {
            Property property;

            if (!_Properties.TryGetValue(name, out property)) return false;

            return true;
        }
       

        /// <summary>
        /// Copyies the contents of this collection to the specified
        /// array.
        /// </summary>
        /// <param name="array">The destination array.</param>
        /// <param name="arrayIndex"></param>
        public void CopyTo(Property[] array, int arrayIndex)
        {
            _Properties.Values.CopyTo(array, arrayIndex);
        }

        /// <summary>
        /// Removes the property with the same name from the collection.
        /// </summary>
        /// <param name="property">The property to remove</param>
        /// <returns>True, if the property was removed. False, otherwise.</returns>
        public bool Remove(Property property)
        {
            return Remove(property.Name);
        }

        /// <summary>
        /// Returns the <see cref="Property"/> with the specified name.
        /// </summary>
        /// <param name="name">The full qualified name of the property</param>
        /// <returns>The <see cref="Property"/> with the specified name.</returns>
        public Property this[string name]
        {
            get
            {
                if (!Contains(name)) throw new PropertyNotFoundException("The requested property was not found");

                return _Properties[name];
            }
        }

        /// <summary>
        /// Returns the <see cref="Property"/> with the specified name.
        /// </summary>
        /// <remarks>The requested property must be a single-valued property</remarks>
        /// <typeparam name="T">The type of the property</typeparam>
        /// <param name="name">The full qualified name of the property</param>  
        /// <returns>The requested property</returns>
        public Property<T> GetProperty<T>(string name)
        {
            return ParseProperty<T>(this[name]);
        }

        private Property<T> ParseProperty<T>(Property property)
        {
            Property<T> result;
            RawProperty rawProperty;

            rawProperty = property as RawProperty;

            if (rawProperty == null) return (Property<T>) property;

            result = rawProperty.GetProperty<T>();
            _Properties[result.Name] = result;

            return result;
        }

        private MultiValuedProperty<T> ParseMultiValuedProperty<T>(Property property)
        {
            MultiValuedProperty<T> result;
            RawProperty rawProperty;

            rawProperty = property as RawProperty;

            if (rawProperty == null) return (MultiValuedProperty<T>)property;

            result = rawProperty.GetMultiValuedProperty<T>();
            _Properties[result.Name] = result;

            return result;
        }


        /// <summary>
        /// Returns the <see cref="Property"/> with the specified name.
        /// </summary>
        /// <remarks>The requested property must be a single-valued property</remarks>
        /// <typeparam name="T">The type of the property</typeparam>
        /// <param name="name">The full qualified name of the property</param>  
        /// <returns>The requested property</returns>
        public MultiValuedProperty<T> GetMultivaluedProperty<T>(string name)
        {
            return ParseMultiValuedProperty<T>(this[name]);
        }
   

        /// <summary>
        /// Tries to get the property with the specified name.
        /// </summary>
        /// <typeparam name="T">The type of the property</typeparam>
        /// <param name="name">The full qualified name of the property</param>  
        /// <param name="property">When the method returns, contains the requested property, if found. Null, otherwise</param>
        /// <returns>True, if the property was found. False, otherwise</returns>
        public bool TryGetProperty<T>(string name, out Property<T> property)
        {
            property = null;

            if (!Contains(name)) return false;

            property = ParseProperty<T>(this[name]);

            return true;
        }

        /// <summary>
        /// Tries to get the multi-valued property with the specified name.
        /// </summary>
        /// <typeparam name="T">The type of the property</typeparam>
        /// <param name="name">The name of the property</param>  
        /// <param name="property">When the method returns, contains the requested property, if found. Null, otherwise</param>
        /// <returns>True, if the property was found. False, otherwise</returns>
        public bool TryGetMultiValuedProperty<T>(string name, out MultiValuedProperty<T> property)
        {
            Property prop;

            property = null;
            if (!Contains(name)) return false;

            prop = _Properties[name];

            property = ParseMultiValuedProperty<T>(prop);

            return true;
        }
     
        /// <summary>
        /// Removes the <see cref="Property"/> with the specified name
        /// from the collection.
        /// </summary>
        /// <param name="name">The full qualified name of the property</param>
        /// <returns>True, if the property was removed. False, otherwise.</returns>
        public bool Remove(string name)
        {
            if (_Properties.Remove(name))
            {
                _RemovedProperties.Add(name);
                return true;
            }

            return false;
        }

         
        /// <summary>
        /// Returns an <see cref="IEnumerator"/> which can be used to iterate
        /// through the collection of properties.
        /// </summary>
        /// <returns>An <see cref="IEnumerator"/> which can be used to iterate
        /// through the collection of properties</returns>
        IEnumerator<Property> IEnumerable<Property>.GetEnumerator()
        {
            return _Properties.Values.GetEnumerator();
        }

    
        #region ICollection<Property> Members

        /// <summary>
        /// Removes all properties from the collection.
        /// </summary>
        public void Clear()
        {
            _RemovedProperties.AddRange(_Properties.Keys);
            _Properties.Clear();
        }

        /// <summary>
        /// Returns the number of properties in the collection.
        /// </summary>
        public int Count
        {
            get { return _Properties.Count; }
        }

        /// <summary>
        /// Returns false;
        /// </summary>
        public bool IsReadOnly
        {
            get { return false; }
        }

        #endregion

        #region IEnumerable Members

        /// <summary>
        /// Returns an <see cref="IEnumerator"/> which can be used to iterate
        /// through the collection of properties.
        /// </summary>
        /// <returns>An <see cref="IEnumerator"/> which can be used to iterate
        /// through the collection of properties</returns>
        public System.Collections.IEnumerator GetEnumerator()
        {
            return _Properties.Values.GetEnumerator();
        }

        #endregion
    }
}