using System;
using System.Collections.Generic;
using System.Text;
using System.DirectoryServices.Protocols;

namespace InfiniTec.DirectoryServices
{
    /// <summary>
    /// Since the System.DirectoryServices.Protocols classes only return 
    /// properties as byte[] or as string values, this class is used to preserve
    /// the original type, until a specific target type is requested.
    /// </summary>
    public class RawProperty: Property
    {
        private DirectoryAttribute _Attribute;

        internal RawProperty(DirectoryAttribute attribute): base(attribute.Name)
        {
            _Attribute = attribute;
        }

        /// <summary>
        /// Gets the property as as typed property.
        /// </summary>
        /// <returns></returns>
        public Property<T> GetProperty<T>()
        {
            string value;

            if (_Attribute.Count == 0) return new Property<T>(_Attribute.Name);

            if (typeof(T) == typeof(byte[]))
            {
                return new Property<T>(_Attribute.Name, (T) _Attribute.GetValues(typeof(byte[]))[0]);
            }
            else if (typeof(T) == typeof(Guid))
            {            
                Guid id;

                id = new Guid((byte[])_Attribute.GetValues(typeof(byte[]))[0]);
                return (Property<T>) (object) new Property<Guid>(_Attribute.Name, id);
            }
            else
            {
                value = (string)_Attribute.GetValues(typeof(string))[0];
                if (typeof(T) == typeof(string)) return new Property<T>(_Attribute.Name, (T)(object)value);
                else return new Property<T>(_Attribute.Name, (T)Convert.ChangeType(value, typeof(T)));
            }
        }

        /// <summary>
        /// Gets the property as multi valued property of the specified type.
        /// </summary>
        /// <returns></returns>
        public MultiValuedProperty<T> GetMultiValuedProperty<T>()
        {
            string[] values;
            T[] typedValues;

            if (_Attribute.Count == 0) return new MultiValuedProperty<T>(_Attribute.Name);

            if (typeof(T) == typeof(byte[]))
            {
                return new MultiValuedProperty<T>(_Attribute.Name, (T[]) (object) _Attribute.GetValues(typeof(byte[])));
            }
            else
            {
                values = (string[])_Attribute.GetValues(typeof(string));
                if (typeof(T) == typeof(string)) return new MultiValuedProperty<T>(_Attribute.Name, (T[]) (object) values);
                else
                {
                    typedValues = Array.ConvertAll<string, T>(values, delegate(string input)
                    {
                        return (T) Convert.ChangeType(input, typeof(T));
                    });
                    return new MultiValuedProperty<T>(_Attribute.Name, (T)Convert.ChangeType(typedValues, typeof(T)));
                }
            }
        }
    }
}
