using System;
using System.Collections.Generic;
using System.Text;
using InfiniTec.Threading;
using System.DirectoryServices.Protocols;
using System.Diagnostics;
using System.ComponentModel;

namespace InfiniTec.DirectoryServices
{

    /// <summary>
    /// This class performs a search on an Active Directory
    /// </summary>
    public class Searcher : AsyncOperationObject
    {
        /// <summary>
        /// Returns a constant which defines the default domain as the search root
        /// for a search operation.
        /// </summary>
        public const string DefaultDomain = "_DefaultDomain";
        /// <summary>
        /// Returns a constant which defines the root domain as the search root
        /// for a search operation.
        /// </summary>
        public const string RootDomain = "_RootDomain";

        /// <summary>
        /// This event is invoked when the progress of the search operation
        /// changed.
        /// </summary>
        public event EventHandler<SearchProgressChangedEventArgs> ProgressChanged;
        /// <summary>
        /// This event is invoked when the the search operation has completed.
        /// </summary>
        public event EventHandler<AsyncCompletedEventArgs> FindCompleted;

        private int _SizeLimit;
        private List<string> _PropertiesToLoad;
        private string _Constraints;
        private SearchScope _Scope;
        private Connection _Connection;
        private string _SearchRoot;
        private List<Item> _Result;
        private int _PageSize;
        private List<SortKey> _SortKeys;
        private NamingContextScope _NamingContextScope;
        private bool _IncludeDeletedItems;
        private SearchToken _SearchToken;

        /// <summary>
        /// Gets or sets the search token.
        /// </summary>
        /// <value>The search token.</value>
        public SearchToken SearchToken
        {
            [DebuggerStepThrough]
            get { return _SearchToken; }
            [DebuggerStepThrough]
            set { _SearchToken = value; }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the search includes deleted items..
        /// </summary>
        /// <value><c>true</c> if deleted items are included in the search; otherwise, <c>false</c>.</value>
        public bool IncludeDeletedItems
        {
            [DebuggerStepThrough]
            get { return _IncludeDeletedItems; }
            [DebuggerStepThrough]
            set { _IncludeDeletedItems = value; }
        }

        /// <summary>
        /// Gets or sets the naming context scope.
        /// </summary>
        /// <value>The naming context scope.</value>
        public NamingContextScope NamingContextScope
        {
            [DebuggerStepThrough]
            get { return _NamingContextScope; }
            [DebuggerStepThrough]
            set { _NamingContextScope = value; }
        }

        /// <summary>
        /// Gets the sort keys.
        /// </summary>
        /// <value>The sort keys.</value>
        public IList<SortKey> SortKeys
        {
            [DebuggerStepThrough]
            get { return _SortKeys; }
        }

        /// <summary>
        /// Gets or sets the size of the page.
        /// </summary>
        /// <value>The size of the page.</value>
        public int PageSize
        {
            [DebuggerStepThrough]
            get { return _PageSize; }
            [DebuggerStepThrough]
            set { _PageSize = value; }
        }

        /// <summary>
        /// Gets the search result.
        /// </summary>
        /// <value>The search result.</value>
        public IList<Item> SearchResult
        {
            [DebuggerStepThrough]
            get { return _Result; }
        }

        /// <summary>
        /// The distinguished name of the directory entry where the search starts.
        /// </summary>
        public string SearchRoot
        {
            [DebuggerStepThrough]
            get { return _SearchRoot; }
            [DebuggerStepThrough]
            set { _SearchRoot = value; }
        }

        /// <summary>
        /// Gets or sets the connection to use.
        /// </summary>
        /// <value>The connection to use.</value>
        public Connection Connection
        {
            [DebuggerStepThrough]
            get { return _Connection; }
            [DebuggerStepThrough]
            set { _Connection = value; }
        }

        /// <summary>
        /// Gets or sets the <see cref="SearchScope"/> of the search operation.
        /// </summary>
        public SearchScope Scope
        {
            get { return _Scope; }
            set { _Scope = value; }
        }

        /// <summary>
        /// Gets or sets the constraints of the operation.
        /// </summary>
        public string Constraints
        {
            get { return _Constraints; }
            set { _Constraints = value; }
        }


        /// <summary>
        /// Provides access to a collection containing the properties to search for.
        /// If this collection is empty, all properties are returned.
        /// </summary>
        public IList<string> PropertiesToLoad
        {
            get { return _PropertiesToLoad; }
        }

        /// <summary>
        /// Gets or sets the maximum number of items returned by the search.
        /// </summary>
        public int SizeLimit
        {
            get { return _SizeLimit; }
            set { _SizeLimit = value; }
        }

        /// <summary>
        /// Calls the <see cref="M:ProgressChanged"/> event.
        /// </summary>
        /// <param name="e"></param>
        protected virtual void OnProgressChanged(SearchProgressChangedEventArgs e)
        {
            if (ProgressChanged != null) ProgressChanged(this, e);
        }

        /// <summary>
        /// Calls the <see cref="M:ProgressCompleted"/> event.
        /// </summary>
        /// <param name="e"></param>
        protected virtual void OnCompleted(AsyncCompletedEventArgs e)
        {
            if (FindCompleted != null) FindCompleted(this, e);
        }

        /// <summary>
        /// Synchronously starts the search operation, retrieving all items.
        /// </summary>
        /// <returns></returns>
        public void FindAll()
        {
            using (AsyncOperation<IList<Item>> op = GetFindOperation(SearchMode.FindAll, false))
            {
                op.Run();
                op.WaitForCompletion();
            }
        }
        /// <summary>
        /// Synchronously starts the search operation, retrieving the next page in the search.
        /// </summary>
        /// <returns></returns>
        public void FindPage()
        {
            using (AsyncOperation<IList<Item>> op = GetFindOperation(SearchMode.FindPage, false))
            {
                op.Run();
                op.WaitForCompletion();
            }
        }
      

        /// <summary>
        /// Asynchronously starts the search operation.
        /// </summary>
        private void FindAsync(SearchMode mode, bool enablePartialResults)
        {
            AsyncOperation<IList<Item>> operation;

            InitiatePendingAsyncOperation();

            operation = GetFindOperation(mode, enablePartialResults);

            RegisterPendingAsyncOperation(operation);

            operation.Completed += SearchAsyncCompleted;
            operation.Error += SearchAsyncError;
            operation.ProgressChanged += SearchProgressChanged;

            operation.Run();
        }

        void SearchProgressChanged(object sender, ProgressChangedEventArgs<IList<Item>> e)
        {
            SearchProgressChangedEventArgs args;

            args = new SearchProgressChangedEventArgs(e.Status);
            _Result.AddRange(args.Items);

            OnProgressChanged(args);

            e.Cancel = args.Cancel;
        }

        void SearchAsyncError(object sender, ExceptionEventArgs e)
        {
            UnregisterPendingAsyncOperation();
            OnCompleted(new AsyncCompletedEventArgs(e.Exception, false, null));
        }

        void SearchAsyncCompleted(object sender, CompletedEventArgs e)
        {
            UnregisterPendingAsyncOperation();

            if (e.Status == OperationStatus.Completed)
            {
                OnCompleted(new AsyncCompletedEventArgs(null, false, null));
            }
            else
            {
                OnCompleted(new AsyncCompletedEventArgs(null, true, null));
            }
        }

        /// <summary>
        /// Returns the search operation used to perform the search based
        /// ojn the specified settings.
        /// </summary>
        /// <returns></returns>
        public virtual AsyncOperation<IList<Item>> GetFindOperation(SearchMode mode, bool enablePartialResults)
        {
            return new AsyncOperation<IList<Item>>(SearchOperationInternal(mode, enablePartialResults));
        }

        private IEnumerable<OperationAction> SearchOperationInternal(SearchMode mode, bool enablePartialResults)
        {
            SearchRequest request;
            SearchResponse response;
            List<Item> result;
            int foundItems;

            do
            {
                if (SearchToken == null && string.IsNullOrEmpty(SearchRoot) && NamingContextScope == NamingContextScope.Domain)
                {
                    throw new ArgumentException("A valid searchroot must be specified when doing a domain-restricted search.");
                }

                if (SearchToken != null && !SearchToken.IsValid())
                {
                    throw new InvalidSearchTokenSpecifiedException("The specified search token is not valid");
                }

                if (SearchToken == null) SearchToken = new SearchToken(PageSize);
                foundItems = 0;

                if (!SearchToken.IsValid() && SearchRoot == DefaultDomain || SearchRoot == RootDomain)
                {
                    Item root;

                    root = new Item(Constants.WellknownDistinguishedNames.RootDse, Connection);

                    if (SearchRoot == DefaultDomain)
                    {
                        root.PropertiesToLoad.Add(Constants.RootDse.DefaultNamingContext);
                    }
                    else
                    {
                        root.PropertiesToLoad.Add(Constants.RootDse.RootNamingContext);
                    }

                    using (InfiniTec.Threading.AsyncOperation operation = root.GetRefreshOperation())
                    {
                        yield return new NestedOperation(operation, false);
                    }

                    if (SearchRoot == DefaultDomain)
                    {
                        SearchRoot = root.Properties.GetProperty<string>(Constants.RootDse.DefaultNamingContext).Value;
                    }
                    else
                    {
                        SearchRoot = root.Properties.GetProperty<string>(Constants.RootDse.RootNamingContext).Value;
                    }
                }

                if (!SearchToken.IsValid())
                {
                    request = new SearchRequest(SearchRoot, Constraints, Scope, _PropertiesToLoad.ToArray());
                }
                else
                {
                    request = new SearchRequest();
                }

                if (SearchToken.IsValid() || PageSize != 0)
                {
                    PageResultRequestControl pagingRequest;

                    pagingRequest = new PageResultRequestControl(PageSize);
                    if (SearchToken.IsValid())
                    {
                        pagingRequest.Cookie = SearchToken.Token;
                        pagingRequest.PageSize = SearchToken.PageSize;
                    }
                    request.Controls.Add(pagingRequest);
                }
            
                if (SortKeys.Count > 0) request.Controls.Add(new SortRequestControl(_SortKeys.ToArray()));
                if (NamingContextScope != NamingContextScope.NotSpecified)
                {
                    switch (NamingContextScope)
                    {
                        case NamingContextScope.Domain:
                            request.Controls.Add(new SearchOptionsControl(SearchOption.DomainScope));
                            break;
                        case NamingContextScope.IncludeSubDomains:
                            request.Controls.Add(new SearchOptionsControl(SearchOption.PhantomRoot));
                            break;
                        default:
                            throw new ArgumentException("NamingContextScope", "The specified value of the NamingContextScope is invalid");
                    }
                }
                if (IncludeDeletedItems) request.Controls.Add(new ShowDeletedControl());

                if (SizeLimit > 0) request.SizeLimit = SizeLimit;

                if (enablePartialResults)
                {
                    using (RequestOperation<IList<Item>, SearchResponse> operation =
                        Connection.GetSendRequestOperation<SearchRequest, SearchResponse, Item>(request,
                        delegate(object entry)
                        {
                            SearchResultEntry item;
                            item = entry as SearchResultEntry;
                            if (item == null) return null;

                            foundItems++;

                            return new Item(item.DistinguishedName, item.Attributes, Connection);
                        }))
                    {
                        yield return new NestedOperation(operation, true);
                        response = operation.Result;
                    }
                }
                else
                {
                    using (RequestOperation<object, SearchResponse> operation =
                        Connection.GetSendRequestOperation<SearchRequest, SearchResponse>(request))
                    {
                        yield return new NestedOperation(operation, true);
                        response = operation.Result;
                    }
                    result = new List<Item>();
                    foundItems = response.Entries.Count;
                    foreach (SearchResultEntry entry in response.Entries)
                    {
                        result.Add(new Item(entry.DistinguishedName, entry.Attributes, Connection));
                    }
                    
                    _Result.AddRange(result);
                    yield return new ProgressUpdate<IList<Item>>(result.AsReadOnly());
                }

                SearchToken.Token = null;
                foreach (DirectoryControl control in response.Controls)
                {
                    PageResultResponseControl prControl;

                    prControl = control as PageResultResponseControl;
                    if (prControl == null) continue;

                    SearchToken.Token = prControl.Cookie;
                }
            } while (mode == SearchMode.FindAll && SearchToken.IsValid());
            if (!SearchToken.IsValid()) SearchToken = null;
        }

        /// <summary>
        /// Creates a new instance of this class using the specified settings.
        /// </summary>
        /// <param name="searchRoot">The root of the search operation.</param>
        /// <param name="connection">The <see cref="Connection"/> to use.</param>
        /// <param name="propertiesToLoad">The properties to load. This parameter is optional.</param>
        public Searcher(string searchRoot, Connection connection, params string[] propertiesToLoad)
            :
            this(connection, propertiesToLoad)
        {
            _SearchRoot = searchRoot;
        }

        /// <summary>
        /// Creates a new instance of this class using the specified settings
        /// </summary>
        /// <param name="connection">The <see cref="Connection"/> to use.</param>
        /// <param name="propertiesToLoad">The properties to load. This parameter is optional.</param>
        public Searcher(Connection connection, params string[] propertiesToLoad)
            : base()
        {
            _Connection = connection;
            _Result = new List<Item>();
            _SortKeys = new List<SortKey>();
            _Scope = SearchScope.Subtree;
            Initialize(propertiesToLoad);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Searcher"/> class.
        /// </summary>
        /// <param name="connection">The connection to use.</param>
        /// <param name="searchToken">The search token to use.</param>
        public Searcher(Connection connection, SearchToken searchToken):
            this(connection)
        {
            SearchToken = searchToken;
        }
	
        private void Initialize(string[] propertiesToLoad)
        {
            _PropertiesToLoad = new List<string>(propertiesToLoad);
        }

        /// <summary>
        /// Asynchronously executes the search operation, finding all items.
        /// </summary>
        public void FindAllAsync()
        {
            FindAsync(SearchMode.FindAll, false);
        }

        /// <summary>
        /// Asynchronously executes the search operation, finding the next page of the search.
        /// </summary>
        public void FindPageAsync()
        {
            FindAsync(SearchMode.FindPage, false);
        }
    }
}