using System;
using System.Collections.Generic;
using System.Text;
using InfiniTec.Threading;
using System.Runtime.InteropServices;
using System.Diagnostics;
using System.Runtime.CompilerServices;

namespace InfiniTec.DirectoryServices
{
    /// <summary>
    /// This class can be used to translate names between different formats, such as
    /// the NT4 account name, the Active-Directory distinguished name, Security identifier,
    /// etc.
    /// </summary>
    public class Translator
    {
        private TranslationFlags _TranslationFlags;
        private NameFormat _InputFormat;
        private NameFormat _OutputFormat;
        private Connection _Connection;
        private TranslationResult[] _Results;

        /// <summary>
        /// Gets the results of the translation process.
        /// </summary>
        /// <value>The results of the translation process.</value>
        public TranslationResult[] Results
        {
            get { return _Results; }
        }

        /// <summary>
        /// Gets or sets the connection to use.
        /// </summary>
        /// <value>The connection to use.</value>
        public Connection Connection
        {
            [DebuggerStepThrough]
            get { return _Connection; }
            [DebuggerStepThrough]
            set { _Connection = value; }
        }

        /// <summary>
        /// Gets or sets the output format.
        /// </summary>
        /// <value>The output format.</value>
        public NameFormat OutputFormat
        {
            [DebuggerStepThrough]
            get { return _OutputFormat; }
            [DebuggerStepThrough]
            set { _OutputFormat = value; }
        }

        /// <summary>
        /// Gets or sets the input format.
        /// </summary>
        /// <value>The input format.</value>
        public NameFormat InputFormat
        {
            [DebuggerStepThrough]
            get { return _InputFormat; }
            [DebuggerStepThrough]
            set { _InputFormat = value; }
        }

        /// <summary>
        /// Gets or sets the translation flags.
        /// </summary>
        /// <value>The translation flags.</value>
        public TranslationFlags TranslationFlags
        {
            [DebuggerStepThrough]
            get { return _TranslationFlags; }
            [DebuggerStepThrough]
            set { _TranslationFlags = value; }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Translator"/> class.
        /// </summary>
        /// <param name="connection">The connection to use.</param>
        public Translator(Connection connection)
        {
            _Connection = connection;
        }


        /// <summary>
        /// Translates the specified names.
        /// </summary>
        /// <param name="names">The names to translate.</param>
        public void Translate(IEnumerable<string> names)
        {
            List<string> list;

            list = new List<string>();
            list.AddRange(names);
            Translate(list.ToArray());
        }
        /// <summary>
        /// Translates the specified names.
        /// </summary>
        /// <param name="names">The names to translate.</param>
        public void Translate(params string[] names)
        {
            uint result;
            IntPtr resultPointer;
            IntPtr pIndex;
            NativeMethods.DS_NAME_RESULT container;
            NativeMethods.DS_NAME_RESULT_ITEM[] resultItems;
            List<TranslationResult> translationResults;

            resultPointer = IntPtr.Zero;

            if (names == null) throw new ArgumentNullException("names", "names must not be null");
            if (names.Length == 0) throw new ArgumentException("names", "At least one entry must be passed to this method");

            if (OutputFormat == NameFormat.SidOrSidHistory) throw new NotSupportedException("The specified nameformat is not supported as the output format");

            RuntimeHelpers.PrepareConstrainedRegions();
            try
            {
                lock (_Connection.NativeConnectionSyncRoot)
                {
                    result = NativeMethods.DsCrackNames(Connection.DirectoryConnection, TranslationFlags,
                        InputFormat, OutputFormat, (uint)names.Length, names, out resultPointer);
                }
                container = (NativeMethods.DS_NAME_RESULT)Marshal.PtrToStructure(resultPointer, typeof(NativeMethods.DS_NAME_RESULT));

                //define the array with size to match				
                resultItems = new NativeMethods.DS_NAME_RESULT_ITEM[container.Count];

                //point to our current DS_NAME_RESULT_ITEM structure
                pIndex = container.Items;

                for (int i = 0; i < container.Count; i++)
                {
                    //marshall back the structure
                    resultItems[i] = (NativeMethods.DS_NAME_RESULT_ITEM)Marshal.PtrToStructure(pIndex, typeof(NativeMethods.DS_NAME_RESULT_ITEM));
                    //update the current pointer idx to next structure
                    pIndex = (IntPtr)(pIndex.ToInt32() + Marshal.SizeOf(resultItems[i]));
                }

                translationResults = new List<TranslationResult>();

                for (int i = 0; i < resultItems.Length; i++)
                {
                    translationResults.Add(
                        new TranslationResult((TranslationStatus)resultItems[i].Status,
                         (i < names.Length) ? names[i] : null, resultItems[i].Name, resultItems[i].Domain));
                }
                _Results = translationResults.ToArray();
            }
            finally
            {
                if (resultPointer != IntPtr.Zero)
                {
                    NativeMethods.DsFreeNameResult(resultPointer);
                }
            }

        }
    }
}
